#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "clear_filesystem_failover"

import sys
import os
import fcntl
import errno

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import failover_common
import lock_common

def clear_filesystem_failover(config_dir):
    # Get lock for job execution
    print_status("Checking if failover is being executed ...")
    execute_lock_path = lock_common.get_execute_lock_file()
    with open(execute_lock_path, 'w') as locker:
        try:
            fcntl.flock(locker, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except EnvironmentError, e:
            if e.errno == errno.EAGAIN:
                print_status_warning("File system synchronization or failover execution is running on this node. Cancel process.")
                raise Exceptions.NoExecution()
            else:
                log_error("An error occurred while getting a lock for failover execution. error=%s" % str(e))
                raise
        else:
            log_info("Execution lock has been acquired.")

    # Delete failover job files, both normal and error
    print_status("Clearing failover jobs ...")
    register_lock_file = lock_common.get_register_lock_file()
    with open(register_lock_file, 'w') as locker:
        # Get registration lock
        fcntl.flock(locker, fcntl.LOCK_EX)

        # Delete pending failover jobs
        while True:
            # Get next job file name
            name = failover_common.get_next_job_name(config_dir)
            if name is None:
                log_info("No available jobs.")
                break

            # Delete the job
            print_status("Deleting failover job for %s ..." % name)
            failover_common.delete_job_file(name, config_dir)

        # Delete success failover jobs
        while True:
            # Get next job file name
            name = failover_common.get_next_job_name(config_dir, is_success=True)
            if name is None:
                log_info("No available jobs.")
                break

            # Delete the job
            print_status("Deleting failover job for %s ..." % name)
            failover_common.delete_job_file(name, config_dir, is_success=True)

        # Delete error failover jobs
        while True:
            # Get next job file name
            name = failover_common.get_next_job_name(config_dir, is_error=True)
            if name is None:
                log_info("No available jobs.")
                break

            # Delete the job
            print_status("Deleting failover job for %s ..." % name)
            failover_common.delete_job_file(name, config_dir, is_error=True)

        # Release registration lock
        fcntl.flock(locker, fcntl.LOCK_UN)

    print_status("All failover jobs has been cleared.")

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    clear_filesystem_failover(config.fo_conf_dir)

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
