#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "failover / failback / dry_run_failover / try_out_failover"

import sys
import os
import getopt
import threading
import time

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HDIClient
import HDIOperation
import FailoverConfig

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import Alert
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import setup_common
import setup_pair_group
import failover_common
import failover_hcp

def clear_failover_jobs(target, config, ssh_config):
    print_status("Clearing failover jobs on %s ..." % target)

    # Prepare connection
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(target), None)
    conn = HDIOperation.Connection(target, cred, config.retry_count, config.retry_interval, config.timeout)

    # Execute 'clear' command on the target node
    prog = os.path.join(BASEDIR, "int/clear_filesystem_failover")
    cmd = "sudo maintexec %s" % prog
    op = HDIOperation.CommandOperation(cmd, display_stdout=True)

    try:
        ret, out, err = op.execute(conn)
    except HDIOperation.ConnectionError, e:
        print_status_error("Failed to communicate with %s. error=%s." % (conn.host, str(e)))
        ret = 1
        # Do not raise an exception to continue failover process for other pairs ...

    if ret != 0:
        log_error("An error occurred when clearing failover jobs on %s. ret=%s" % (conn.host, ret))

def register_failover_job(pair_id, source, target, fs, config, ssh_config, policy=None, is_failback=False):
    # Prepare connection
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(target), None)
    conn = HDIOperation.Connection(target, cred, config.retry_count, config.retry_interval, config.timeout)

    # Execute 'register' command on the target node
    prog = os.path.join(BASEDIR, "int/register_filesystem_failover")
    failback_option = "--failback" if is_failback else ""
    if policy is None:
        cmd = "sudo maintexec %s --id %s --source %s --file-system %s %s" % (prog, pair_id, source, fs, failback_option)
    else:
        cmd = "sudo maintexec %s --id %s --source %s --file-system %s --storage-policy %s %s" % (prog, pair_id, source, fs, policy, failback_option)
    op = HDIOperation.CommandOperation(cmd, display_stdout=True)

    try:
        ret, out, err = op.execute(conn)
    except HDIOperation.ConnectionError, e:
        print_status_error("[ID: %s] Failed to communicate with %s. error=%s" % (pair_id, conn.host, str(e)))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if ret != 0:
        log_error("[ID: %s] Command finished with an error. host=%s, cmd=\"%s\", ret=%s" % (pair_id, conn.host, cmd, ret))
        print_status_error("[ID: %s] Failover job registration command finished with an error on %s." % (pair_id, conn.host))

    return ret

def execute_failover_job(pair_id, target, config, ssh_config, is_retry_mode, rehearsal=None):
    # Prepare connection
    cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(target), None)
    conn = HDIOperation.Connection(target, cred, config.retry_count, config.retry_interval, config.timeout)

    # Execute 'execute' command on the target node
    options = ""
    if is_retry_mode:
        options += "--error "
    if rehearsal == failover_common.RehearsalMode.DRYRUN:
        options += "--dry-run "
    elif rehearsal == failover_common.RehearsalMode.TRYOUT:
        options += "--try-out "

    prog = os.path.join(BASEDIR, "int/execute_filesystem_failover --id %s %s" % (pair_id, options))
    cmd = "sudo maintexec %s" % prog
    op = HDIOperation.CommandOperation(cmd, display_stdout=True)

    try:
        ret, out, err = op.execute(conn)
    except HDIOperation.ConnectionError, e:
        print_status_error("[ID: %s] Failed to communicate with %s. error=%s" % (pair_id, conn.host, str(e)))
        return 1

    if ret != 0:
        log_error("[ID: %s] Command finished with an error. host=%s, cmd=\"%s\", ret=%s" % (pair_id, conn.host, cmd, ret))
        print_status_error("[ID: %s] Failover job execution command finished with an error on %s." % (pair_id, conn.host))
        return ret

    return 0

def failover(pair_id, pairconf, stconf, config, ssh_config, is_retry_mode, retcodes, is_failback=False, rehearsal=None):
    # Prepare information
    pair_list = pairconf.get_list(pair_id)
    notused, pair = pair_list[0]
    if is_failback:
        source = pair.secondary
        target = pair.primary
        operation = 'failback'
    else:
        source = pair.primary
        target = pair.secondary
        operation = 'failover'

    print_status("[ID: %s] ===== %s start from %s to %s. =====" % (pair_id, operation.capitalize(), source, target))

    # Register failover job on the target node
    # First execution only
    error = 0
    if not is_retry_mode:
        if pair.is_all:
            print_status("[ID: %s] Registering a failover job on %s for all file systems ..." % (pair_id, target))
            try:
                ret = register_failover_job(pair_id, source, target, "\<ALL\>", config, ssh_config, is_failback=is_failback)
            except Exceptions.ConnectionError, e:
                if rehearsal != failover_common.RehearsalMode.DRYRUN:
                    # Send an alert here because we couldn't connect the target HDI at all.
                    # Do not send alert in dry-run failover case.
                    Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname="<ALL>", detail=str(e))
                ret = 1
            if ret != 0:
                print_status_error("[ID: %s] Registering a failover job on %s failed." % (pair_id, target))
                # Continue to execute failover file systems as many as possible.
                # It is possible that some of the file systems are already registered
                # even if register function itself returned with an error.
                # An alert should be raised on the target node.
                error = ret
        else:
            for fs in pair.target_fslist:
                print_status("[ID: %s] Registering a failover job on %s for file system %s ..." % (pair_id, target, fs))
                try:
                    ret = register_failover_job(pair_id, source, target, fs, config, ssh_config, is_failback=is_failback)
                except Exceptions.ConnectionError, e:
                    if rehearsal != failover_common.RehearsalMode.DRYRUN:
                        # Send an alert here because we couldn't connect the target HDI at all.
                        Alert.send_alert(Alert.MsgCode.FAILOVER_ERROR, fsname=fs, detail=str(e))
                    ret = 1
                if ret != 0:
                    print_status_error("[ID: %s] Registering a failover job on %s failed. Continue processing the remaining file systems ..." % (pair_id, target))
                    # Continue to execute failover file systems as many as possible.
                    # Record error code to return it later
                    error = ret
    else:
        print_status("[ID: %s] Skip registration of failover jobs because this is 'retry' execution." % pair_id)

    # Execute failover jobs on the target node
    if is_retry_mode:
        print_status("[ID: %s] Executing failover jobs on %s as retry mode ..." % (pair_id, target))
    else:
        print_status("[ID: %s] Executing failover jobs on %s ..." % (pair_id, target))
    ret = execute_failover_job(pair_id, target, config, ssh_config, is_retry_mode, rehearsal=rehearsal)
    if ret != 0:
        print_status_error("[ID: %s] Executing failover jobs on %s failed." % (pair_id, target))
        retcodes[pair_id] = ret
    else:
        retcodes[pair_id] = error

    print_status("[ID: %s] ===== %s finished from %s to %s. =====" % (pair_id, operation.capitalize(), source, target))
    return

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    short_opts = ''
    long_opts = ['failback', 'dry-run', 'try-out']
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    is_failback = False
    rehearsal = None
    for o, a in opts:
        if o == '--failback':
            is_failback = True
        if o == '--dry-run':
            rehearsal = failover_common.RehearsalMode(failover_common.RehearsalMode.DRYRUN)
        if o == '--try-out':
            rehearsal = failover_common.RehearsalMode(failover_common.RehearsalMode.TRYOUT)

    operation = 'failback' if is_failback else 'failover'

    # Load the existing failover configuration
    log_info("Getting list of failover configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Select a configuration group
    confgroup = setup_common.select_configuration_group(confgroup_list)

    # Load pair configuration and pair group configuration
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
    pgconf = FailoverConfig.PairGroupConfig(config.fo_conf_dir, confgroup)
    stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, confgroup)

    # Setup SSHConfig for the configuration group
    ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

    ### Select a pair group to execute failover if it is requested
    # Display defined pair groups
    print
    print "----- Defined pair groups -----"
    setup_pair_group.display_pair_groups(pgconf, pairconf)

    # Use pair group as failover unit?
    if len(pgconf.get_list()) != 0:
        use_pair_group = Utility.yes_or_no("Do you want to execute %s per pair group?" % operation)
    else:
        use_pair_group = False

    if use_pair_group:
        # Select a pair group to failover
        print
        print ">>> Select a pair group to %s." % operation
        name = setup_pair_group.select_pair_group(pgconf)

        # Get a list of pairs that belong to the pair group
        pg_list = pgconf.get_list(name=name)
        notused, pairgroup = pg_list[0]
        pid_list = pairgroup.pair_list
    else:
        # Select a pair(s) to failover
        print
        print ">>> Select a pair(s) to %s." % operation
        pid_list = setup_common.select_pair(pairconf, msg="Select a pair(s) (separate with comma for multiple selection)", onlyone=False)
        if pid_list is None:
            print
            print "No pairs are selected."
            log_info("No pairs are selected.")
            raise Exceptions.NoExecution()

    ### First execution or retry
    print
    print ">>> Select execution mode."
    MODE_FIRST = 1
    MODE_RETRY = 2
    mode_menu = [
        (MODE_FIRST, "First execution"),
        (MODE_RETRY, "Second execution to retry failed %s" % operation)
    ]
    selected_list = Utility.select_items("Select execution mode", mode_menu, title="Execution mode", onlyone=True)
    if selected_list is None:
        log_warning("Selecting an execution mode did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()

    is_retry_mode = selected_list[0] == MODE_RETRY

    ### HCP failover
    do_hcp_failover = False
    do_change_primary_hcp_on_hdi = False
    if rehearsal != failover_common.RehearsalMode.DRYRUN:
        # HCP failover option is selectable only if try-out failover or normal failover
        print
        print ">>> Select HCP %s option." % operation
        if not is_failback:
            # Select if HCP failover is executed or not
            HCP_FAILOVER_AND_SWAP = 1
            HCP_FAILOVER_ONLY = 2
            HCP_NONE = 0
            hcp_menu = [
                (HCP_FAILOVER_AND_SWAP, "Execute HCP failover AND change primary HCP setting on HDI"),
                (HCP_FAILOVER_ONLY, "Execute HCP failover"),
                (HCP_NONE, "Do nothing about HCP failover")
            ]
        else:
            # Select if HCP failback has been executed and change primary HCP on HDI
            HCP_FAILBACK_SWAP = 1
            HCP_NONE = 0
            hcp_menu = [
                (HCP_FAILBACK_SWAP, "(Only if HCP failback is complete) Change primary HCP setting on HDI"),
                (HCP_NONE, "Do nothing about HCP failback")
            ]
        selected_list = Utility.select_items("Select HCP %s operation" % operation, hcp_menu, title="HCP %s operations" % operation, onlyone=True)
        if selected_list is None:
            log_warning("Selecting an HCP %s operation did not succeed." % operation)
            print "Process canceled."
            raise Exceptions.NoExecution()
        selected = selected_list[0]
        if is_failback:
            if selected == HCP_FAILBACK_SWAP:
                do_change_primary_hcp_on_hdi = True
        else:
            if selected == HCP_FAILOVER_AND_SWAP:
                do_hcp_failover = True
                do_change_primary_hcp_on_hdi = True
            elif selected == HCP_FAILOVER_ONLY:
                do_hcp_failover = True

    ### Confirmation
    # Display selected pairs for confirmation
    print
    print "----- %s configuration -----" % operation.capitalize()
    if use_pair_group:
        print
        print "Pair group: %s" % name
        setup_pair_group.display_pairs_in_pair_group(name, pgconf, pairconf)
    else:
        print
        print "Pairs:"
        for pid in pid_list:
            pair_list = pairconf.get_list(pair_id=pid)
            notused, pair = pair_list[0]
            print "\t%s - %s" % (pair.primary, pair.secondary)
    print
    print "Operation:"
    print "\t%s" % operation.capitalize()
    print
    print "Rehearsal:"
    print "\t%s" % ("No" if rehearsal is None else rehearsal)
    print
    print "HCP failover:"
    if do_hcp_failover:
        print "\tExecute HCP %s" % operation
    if do_change_primary_hcp_on_hdi:
        print "\tChange primary HCP setting on HDI"
    if not do_hcp_failover and not do_change_primary_hcp_on_hdi:
        print "\tNothing"
    print
    print "Execution mode:"
    print "\t%s" % ("Second execution to retry failed %s" % operation if is_retry_mode else "First execution")
    print "--------------------------------------"
    print

    # Confirmation question
    answer = Utility.yes_or_no_full("Do you want to execute %s with the configuration above?" % operation)
    if answer is False:
        # Canceled
        raise Exceptions.NoExecution()

    ### Execute HCP failover / failback operation
    if do_hcp_failover:
        # Execute HCP failover
        print_status("Execute HCP failover.")
        failover_hcp.failover_hcp(config)

    if do_change_primary_hcp_on_hdi:
        # Change HCP setting on each HDI
        # Pick up HDI nodes to change HCP setting.
        # If this is failover, select only secondary nodes that are included in the selected pairs.
        # If this is failback, select all nodes that are included in the selected pairs.
        print_status("Change primary HCP setting on HDIs.")
        nodes = []
        for pid in pid_list:
            pair_list = pairconf.get_list(pair_id=pid)
            notused, pair = pair_list[0]
            if is_failback:
                nodes.append(pair.primary)
                nodes.append(pair.secondary)
            else:
                nodes.append(pair.secondary)
        if is_failback:
            print_status("All nodes in the selected pairs will be updated as this is failback.")
        else:
            print_status("Only secondary nodes in the selected pairs will be updated as this is failover.")
        failover_hcp.change_hcp_on_hdi(config, nodes=nodes)

    ### Execute HDI failover
    # Clear failover jobs on the target nodes if this is "First execution"
    print_status("=== Preparation ===")
    if not is_retry_mode:
        for pid in pid_list:
            pair_list = pairconf.get_list(pid)
            notused, pair = pair_list[0]
            if is_failback:
                target = pair.primary
            else:
                target = pair.secondary
            clear_failover_jobs(target, config, ssh_config)

    # Execute failover for each pair in parallel
    start = int(time.time())
    thread_list = []
    retcodes = {}
    for pid in pid_list:
        t = threading.Thread(target=failover, args=(pid, pairconf, stconf, config, ssh_config, is_retry_mode, retcodes, is_failback, rehearsal))
        t.start()
        thread_list.append(t)

    # Wait for all execution finish
    for t in thread_list:
        t.join()

    # Get resulsts
    print_status("%s for all pairs finished." % operation.capitalize())
    for pid in sorted(pid_list):
        l = pairconf.get_list(pid)
        pair = l[0][1]
        if is_failback:
            source = pair.secondary
            target = pair.primary
        else:
            source = pair.primary
            target = pair.secondary
        if pid not in retcodes.keys():
            print_status_warning("No return code for pair ID %s (from %s to %s) for some reason." % (pid, source, target))
        elif retcodes[pid] == 0:
            print_status("Pair ID %s (from %s to %s) finished normally." % (pid, source, target))
        else:
            print_status_error("Pair ID %s (from %s to %s) finished with an error. Check logs and messages on each node to see the current status." % (pid, source, target))

    # Collect and display result of failover time
    print_status("Collecting failover results ...")
    results = {}
    for pid in sorted(pid_list):
        l = pairconf.get_list(pid)
        pair = l[0][1]
        if is_failback:
            target = pair.primary
        else:
            target = pair.secondary
        # The same target node can appear in N to 1 case
        if target not in results.keys():
            try:
                print_status("Getting failover result from %s ..." % target)
                results[target] = failover_common.get_result(target, config.fo_conf_dir, config, ssh_config)
                print_status("Initializing failover result on %s ..." % target)
                failover_common.initialize_remote_result(target, config, ssh_config)
            except (Exceptions.ConnectionError, Exceptions.OperationError), e:
                # Give up to display the result if failed to get...
                log_error("Failed to get or initialize failover result on %s. error=%s. Continue anyway ..." % (target, str(e)))
                # Pass through

    # Display result
    print_status("Display failover results.")
    failover_common.display_result(results, start, rehearsal)

    # Instruct to execute complete_xxxx
    if rehearsal is None:
        print
        print "Please do the followings:"
        print "\t1. Confirm all configurations are restored as expected."
        print "\t2. If there are failed file systems, try %s again with:" % operation
        print "\t\t- Specifying the exact pair(s) that failed to %s" % operation
        print "\t\t- Speicfying 'Second execution' mode"
        print "\t3. Execute complete_%s script on *each* %s node(s) to complete the %s." % (operation, "primary" if is_failback else "secondary", operation)
        print

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
