#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"

import sys
import os

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import FailoverConfig
import HDIClient
import HDIOperation
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception
import Utility

sys.path.append(os.path.join(BASEDIR, "bin"))
import Exceptions

# Functions

def select_operation(menu):
    print
    print ">>> Select an operation."
    selected = Utility.select_items("Select an operation.", menu, title="Operations", onlyone=True)
    if selected is None:
        log_warning("Selecting an operation did not succeed.")
        print "Process canceled."
        raise Exceptions.NoExecution()

    return selected[0]

def select_configuration_group(confgroup_list):
    if len(confgroup_list) == 0:
        # Define a new configuration group
        print "*** There is no configuration group yet. ***"
        print
        print ">>> Define a new configuration group name."
        print ">>> Please use only alphanumeric characters and underscore, not more than 20 characters."
        print
        name = Utility.my_input("New configuration group name: ")
        if name is None or not Utility.is_word(name) or len(name) > 20:
            log_warning("Nothing was input or a wrong name was input for configuration group name. name=%s" % name)
            print "Invalid input for configuration group name."
            raise Exceptions.NoExecution()
        elif name == 'default':
            log_info("'default' is entered as a configuration group name.")
            print "Sorry, 'default' cannot be used as a configuration group name. Please use another name."
            raise Exceptions.NoExecution()
        else:
            return name
    elif len(confgroup_list) == 1:
        return confgroup_list[0]
    else:
        print "There are more than 1 configuration groups."
        print
        print ">>> Select a configuration group that you want to operate."
        menu = []
        for idx, confgroup in enumerate(confgroup_list):
            menu.append((idx + 1, confgroup))
        selected = Utility.select_items("Select a configuration group.", menu, title=None, onlyone=True)
        if selected is None:
            log_warning("Selecting a configuration group did not succeed.")
            print "Process canceled."
            raise Exceptions.NoExecution()

    return confgroup_list[selected[0] - 1]

def display_pair_line(idx, primary, secondary, filesystem, storage_policy):
    if idx is None:
        idx = ''
    if primary is None:
        primary = ''
    if secondary is None:
        secondary = ''

    # idx primary secondary filesystem policy
    print "%-5s%-16s%-16s%-17s%s" % (idx, primary, secondary, filesystem, storage_policy)

def display_pair(idx, pair, policy_list):
    # policy_list is a list of tuples of (pair ID, file system, storage policy)
    # for the pair (i.e. pair ID is same for all elements in policy_list)

    log_debug("Displaying a pair. pair: %s, policy_list: %s" % (pair, Utility.deep_str(policy_list)))
    # Get default policy
    def_policy_list = [policy for policy in policy_list if policy[1] == FailoverConfig.StoragePolicy.FSDEF]
    if len(def_policy_list) == 0:
        log_error("No default storage policy is defined. Default storage policy is mandatory. policy_list: %s" % policy_list)
        raise Exceptions.ConfigurationError("No default storage policy is defined for some reason.")
    default_policy = def_policy_list[0][2]

    # Build pairs of target file system and policy
    fs_policy_list = []
    if pair.is_all:
        # All file systems are target of failover.
        # Default policy is applied to all file systems unless a specific policy
        # is assigned to individual file system.

        # Default policy is applied to all file systems
        is_default = True
        fs_policy_list.append(("(All)", default_policy, is_default))

        # Display individual policies if defined
        is_default = False
        ind_policy_list = [policy for policy in policy_list if policy[1] != FailoverConfig.StoragePolicy.FSDEF]
        if len(ind_policy_list) != 0:
            for notused, fs, policy in ind_policy_list:
                fs_policy_list.append((fs, policy, is_default))
    elif pair.target_fslist is None:
        # No file systems are target of failover
        fs_policy_list.append(("(None)", None, False))
    else:
        # A part of file systems are target of failover
        for fs in pair.target_fslist:
            ind_policy_list = [policy for policy in policy_list if policy[1] == fs]
            if len(ind_policy_list) == 0:
                # No policy is defined for this file system. Display it as default policy is applied.
                is_default = True
                fs_policy_list.append((fs, default_policy, is_default))
            else:
                # Expect only one policy for one file system defined
                is_default = False
                notused, fs, policy = ind_policy_list[0]
                fs_policy_list.append((fs, policy, is_default))

    log_debug("fs_policy_list: %s" % Utility.deep_str(fs_policy_list))

    # Display
    first = True
    for fs, policy, is_default in fs_policy_list:
        if policy is None:
            policy_string = '-'
        else:
            policy_string = policy.to_display_string()

        if is_default:
            policy_string += ' (default)'

        if first:
            display_pair_line(idx, pair.primary, pair.secondary, fs, policy_string)
            first = False
        else:
            display_pair_line(None, None, None, fs, policy_string)

def display_pairs(pairconf, stconf):
    """ Display defined pairs and storage policies.
    pairconf: PairConfig
    stconf: StoragePolicyConfig
    """

    print
    pair_list = pairconf.get_list()
    if len(pair_list) == 0:
        print "*** There is no pairs defined. ***"
    else:
        # Display title
        display_pair_line('#', 'Primary', 'Secondary', 'File system', 'Storage policy')
        print "--------------------------------------------------------------------------"
        idx = 0
        for pid, pair in pair_list:
            idx += 1
            policy_list = stconf.get_list(pair_id=pid)
            display_pair(idx, pair, policy_list)
            print "--------------------------------------------------------------------------"
    print
    print "Finished."

def select_pair(pairconf, msg=None, onlyone=True):
    """ Select a pair(s) from defined pairs, and return pair ID of the
    selected pair(s).

    This function provides a common dialog to select a pair or pairs
    from the defined pairs.
    Message to be displayed when selecting a pair can be specified as
    'msg' argument.
    If onlyone=True is specified, only one pair can be selected and
    the pair ID is returned. Otherwise, multiple pairs can be selected
    and a list of pair ID is returned in that case.

    Args:
        pairconf (FailoverConfig.PairConfig): Pair configuration.
        msg (str): Message to be displayed when selecting a pair.
        onlyone (bool): Specify True if only one pair should be selected.
    Returns:
        int or list of int: A pair ID is returned when onlyone=True is
            specified. Otherwise a list of pair IDs is returned.
        None is returned if no pairs are selected.
    """
    if msg is None:
        msg = "Select a pair to operate"

    # Get the current defined pairs
    pair_list = pairconf.get_list()
    log_debug("Pair list: %s" % Utility.deep_str(pair_list))
    if len(pair_list) == 0:
        print
        print "*** No pairs are defined. ***"
        log_info("No pairs are defined.")
        return None

    # Select a pair
    pair_menu = []
    pair_id_map = {}
    for i, (pid, pair) in enumerate(pair_list):
        idx = i + 1
        pair_menu.append((idx, "%s - %s" % (pair.primary, pair.secondary)))
        pair_id_map[idx] = pid
    selected_list = Utility.select_items(msg, pair_menu, title="Defined pairs", onlyone=onlyone)
    if selected_list is None:
        log_warning("Selecting a pair did not succeed.")
        return None

    if onlyone:
        selected = selected_list[0]
        pid = pair_id_map[selected]
        return pid
    else:
        pid_list = []
        for idx in selected_list:
            pid_list.append(pair_id_map[idx])
        return pid_list

def get_my_ip_address(config, confgroup=None):
    """ Find an IP address of the node from pair configuration and returns it.

    This function is used to know which IP address of the node is defined in
    pair configuration.
    Also the IP address is used as an identifier of the node in this script set.

    This script gets a list of IP addresses of this node at first, then compares
    it with a list of nodes defined in pair configuration.

    Args:
        config (Config.Config): Configuration object of the scripts.
        confgroup (str): Configuration group name of the pair configuration.
    Returns:
        str: IP address of the node which is used in the pair configuration.
    """
    log_info("Getting IP address of this node ...")
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)
    netop = HDIOperation.Network(conn)
    try:
        iplist = netop.get_iplist()
    except HDIOperation.OperationError, e:
        log_error("Failed to get IP address of this node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get IP address of this node. reason=%s, code=%s" % (e.reason, e.code), e.code)
    log_info("IP addresses of this node: %s" % iplist)

    # Find my IP address in the node list. I'm not sure which IP address
    # is used in pair configuration.
    self_ip = FailoverConfig.PairConfig.find_matched_ip(config.fo_conf_dir, iplist, confgroup)
    log_info("My IP address in the pair node list: %s" % self_ip)

    return self_ip

def is_primary(node, config_dir, confgroup=None):
    """ Check if a node is defined as a primary node or not.

    This function enumerates all pairs defined in the specified configuration
    directory and checks if the specified node is defined as a primary node
    or a secondary node.
    Returns True if the node is defined as a primary node. Otherwise returns
    False.

    If configuration group is provided, this function checks only the pair configuration of
    the configuration group.

    Args:
        node (str): IP address of a node that you want to check.
        config_dir (str): Directory path where pair configuration is stored.
        confgroup (str): Configuration group name of pair configuration that you want to find the node
            from.
    Returns:
        bool: True if the node is defined as a primary node. False otherwise.
            None can be returned if the specified node is not found in pair
            configuration or None is specified as 'node'.
    """

    log_info("Checking if %s is primary or not (group=%s)..." % (node, confgroup))

    if node is None:
        return None

    result = None
    if confgroup is None:
        confgroup_list = FailoverConfig.list_configuration_group(config_dir)
    else:
        confgroup_list = [confgroup]

    for confgroup in confgroup_list:
        pairconf = FailoverConfig.PairConfig(config_dir, confgroup)
        # Try to get pairs where the node is defined as primary
        node_list = pairconf.get_list(primary=node)
        if len(node_list) != 0:
            # Found. The node is primary.
            result = True
            log_info("%s is defined as a primary node." % node)
            break
        # Try to get pairs where the node is defined as secondary
        node_list = pairconf.get_list(secondary=node)
        if len(node_list) != 0:
            # Found. The node is secondary.
            result = False
            log_info("%s is defined as a secondary node." % node)
            break
    else:
        log_info("%s is not found in the pair configuration." % node)

    return result
