#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "sync_fs"

import sys
import os
import fcntl
import errno
import getopt

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HDIClient
import HDIConfig
import HDIOperation
import FailoverConfig

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import Alert
import HDIConfigArchive
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import setup_common
import getinfo_common
import backup_sync_common
import lock_common
import check_resources

send_alert = False
got_error = False

def sync_filesystem(fsinfo, partner, policy, config, fslist_on_this_node=None, original_fsinfo=None):
    """ Synchronize file system configuration.

    Synchronize a file system with the partner's configuration.
    "Synchronize" means:
        Create the file system if not yet.
        Re-create if any configuration of the file system or the storage
        policy has been changed on the partner node.

    If fslist_on_this_node is provided, this function checks the existence
    of the specified file system in the list. Otherwise this function issues
    MAPI to get the existence of the file system which takes longer time.

    After synchronized the file system, this function saves the current
    configuration of the file system to "synced" directory to use it at
    the next synchronization for configuration comparison.

    Args:
        fsinfo (HDIConfig.FileSystemInfo): File system information.
        partner (str): The IP address of the partner where the file system
            is failed over from.
        policy (FailoverConfig.StoragePolicy): Storage configuration policy
            for the file system.
        config (Config.Config): Configuration object of the scripts.
        fslist_on_this_node (list of str): List of file system name on this
            node.
    Returns:
        None
    """
    # Prepare file system operation
    # Setup connection to localhost
    # SSH credential is not required here because this is connection to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection("localhost", cred, config.retry_count, config.retry_interval, config.timeout)
    fsop = HDIOperation.FileSystem(conn)

    # Is this file system supportd?
    if not fsinfo.is_supported():
        print_status("Skip %s because it is not supported." % fsinfo.fileSystemName)
        return

    # Is this already created?
    if fslist_on_this_node is None:
        try:
            created = fsop.exists(fsinfo.fileSystemName)
        except HDIOperation.OperationError, e:
            log_error("Failed to get existence of %s on %s. error=%s" % (fsinfo.fileSystemName, conn.host, str(e)))
            raise Exceptions.OperationError("Failed to get existence of a file system. error=%s" % e.reason, e.code)
    else:
        created = fsinfo.fileSystemName in [e.fileSystemName for e in fslist_on_this_node]

    # Decide if this file system needs to be created/re-created
    if created:
        # Already created.
        print_status("%s already exists." % fsinfo.fileSystemName)

        # Get file system information and storage policy at the last synchronized time if it exists
        synced_fsinfo = backup_sync_common.get_synced_fsinfo(fsinfo.fileSystemName, partner, config.hdi_conf_dir)
        synced_policy_string = backup_sync_common.get_synced_policy(fsinfo.fileSystemName, partner, config.hdi_conf_dir)
        need_to_create = False

        if synced_fsinfo is None:
            # No synced information. Recreate it anyway.
            print_status("No synced configuration found for %s. Recreate anyway." % fsinfo.fileSystemName)
            need_to_create = True
        else:
            # Check if synced storage policy exists
            if synced_policy_string is None:
                # Recreate anyway the file system if we couldn't find a policy at the last sync
                print_status("No synced storage policy found for %s. Recreate anyway." % fsinfo.fileSystemName)
                need_to_create = True

            # Check if storage policy has been changed
            elif policy.to_string() != synced_policy_string:
                # Storage policy for this file system has been changed. Need to recreate.
                # Create a new StoragePolicy instance just for display the policy to console.
                old_policy = FailoverConfig.StoragePolicy(fsinfo.fileSystemName, policy_string=synced_policy_string)
                print_status("Storage policy for %s has been changed (old: %s, new: %s). Need to recreate." % (fsinfo.fileSystemName, old_policy.to_display_string(), policy.to_display_string()))
                need_to_create = True

            # Compares configuration
            elif fsinfo.has_different_configuration_from(synced_fsinfo):
                # Some configuration has been changed. Need to recreate.
                print_status("Configuration of %s has been changed. Need to recreate." % fsinfo.fileSystemName)
                need_to_create = True

            # Check if the remaining capacity is enough for "Auto - usage" policy
            elif policy.policy == FailoverConfig.StoragePolicy.AUTOUSAGE:
                print_status("Storage policy of %s is \"%s\". Checking capacity ..." % (fsinfo.fileSystemName, policy.to_display_string()))
                # Get capacity of the file system on this node
                try:
                    local_fsinfo = fsop.get(fsinfo.fileSystemName)
                except HDIOperation.OperationError, e:
                    log_error("Failed to get file system information of %s. error=%s" % (fsinfo.fileSystemName, str(e)))
                    raise Exceptions.OperationError("Failed to get file system information. error=%s" % e.reason, e.code)
                if local_fsinfo is None:
                    # This should not happen because we already checked the existence of this file system.
                    log_warning("File system %s does not exist? Recreate anyway ..." % fsinfo.fileSystemName)
                    need_to_create = True
                else:
                    # Compare capacity of the file system on this node and usage of the file system on the pair node
                    if not has_enough_capacity(local_fsinfo, fsinfo, policy.margin_capacity):
                        # Not enough capacity. Need to recreate.
                        print_status("Capacity of %s on this node is not enough (usage on remote: %s GB, capacity on local: %s GB). Need to recreate." % (fsinfo.fileSystemName, int(fsinfo.usedBlocksInMegaBytes / 1024), local_fsinfo.fileSystemSizeInGigaByte))
                        need_to_create = True
                    else:
                        # Still has enough capacity.
                        print_status("Capacity of %s on this node is still enough." % fsinfo.fileSystemName)
                        need_to_create = False
            else:
                # No changes on file system setttings. Don't need to re-create.
                print_status("Configuration of %s has not changed." % fsinfo.fileSystemName)
                need_to_create = False
    else:
        # Not created yet.
        print_status("%s does not exist." % fsinfo.fileSystemName)
        need_to_create = True

    if need_to_create:
        if created:
            # Need to delete it at first
            try:
                # Unmount
                print_status("Checking mount status of %s ..." % fsinfo.fileSystemName)
                is_mounted = fsop.is_mounted(fsinfo.fileSystemName)
                if is_mounted:
                    print_status("Unmounting %s ..." % fsinfo.fileSystemName)
                    fsop.unmount(fsinfo.fileSystemName)
                else:
                    print_status("%s is already unmounted." % fsinfo.fileSystemName)
                # Delete
                print_status("Deleting %s ..." % fsinfo.fileSystemName)
                fsop.delete(fsinfo.fileSystemName)
            except HDIOperation.OperationError, e:
                print_status_error("Failed to unmount/delete a file system %s on %s. error=%s" % (fsinfo.fileSystemName, conn.host, str(e)))
                # Delete an existing file system info file to recreate this file system
                # forcibly at the next synchronization.
                backup_sync_common.delete_fsinfo(fsinfo.fileSystemName, partner, config.hdi_conf_dir)
                raise Exceptions.OperationError("Failed to delete a file system. error=%s" % e.reason, e.code)

        # Let's create the file system!
        try:
            # Create
            print_status("Creating %s ..." % fsinfo.fileSystemName)
            fsop.create(fsinfo, policy, original_fsinfo)
            # Mount
            print_status("Mounting %s ..." % fsinfo.fileSystemName)
            fsop.mount(fsinfo.fileSystemName, mountinfo=fsinfo.MountSetting)
        except HDIOperation.UnsupportedError, e:
            print_status_warning("Tried to create/mount an unsupported file system %s on %s. error=%s" % (fsinfo.fileSystemName, conn.host, str(e)))
            # Just skip this file system
            return
        except HDIOperation.OperationError, e:
            print_status_error("Failed to create a file system %s on %s. error=%s" % (fsinfo.fileSystemName, conn.host, str(e)))
            # Delete an existing file system info file to recreate this file system
            # forcibly at the next synchronization.
            backup_sync_common.delete_fsinfo(fsinfo.fileSystemName, partner, config.hdi_conf_dir)
            raise Exceptions.OperationError("Failed to create a file system. error=%s" % e.reason, e.code)
    # Save the file system information
    print_status("Saving the latest configuration ...")
    backup_sync_common.save_fsinfo(fsinfo, partner, config.hdi_conf_dir)
    backup_sync_common.save_policy(policy.to_string(), fsinfo.fileSystemName, partner, config.hdi_conf_dir)

def has_enough_capacity(local_fsinfo, remote_fsinfo, margin_capacity):
    """ Decide if the local file system has enough capacity to restore data
    on the remote file system based on the margin capacity specified by the
    storage policy.

    This function is used to decide if we need to recreate a file system
    when "Auto - usage" storage policy is applied.
    This function decides "enough" if the local capacity is larger than the
    remote usage + margin / 2. This means the local file system should always
    have capacity more than the usage of the remote file system + margin / 2.

    Example:
        local capacity = 500GB, remote usage = 400GB, margin = 100GB
        -> Enough.
        local capacity = 500GB, remote usage = 430GB, margin = 100GB
        -> Enough.
        local capacity = 500GB, remote usage = 460GB, margin = 100GB
        -> Not enough. remote usage + margin / 2 exceeds local capacity.

    The reason to use "margin / 2", not just "margin", is to have robustness
    against usage increase on the remote file system.

    Args:
        local_fsinfo (HDIConfig.FileSystem): File system information object
            of the local file system.
        remote_fsinfo (HDIConfig.FileSystem): File system information object
            of the remote file system.
        margin_capacity (int): Margin capacity specified by the storage policy.
    Returns:
        bool: True if it has enough capacity. False otherwise.
    """

    local_capacity = local_fsinfo.fileSystemSizeInGigaByte
    remote_usage = int(remote_fsinfo.usedBlocksInMegaBytes / 1024)
    log_info("Capacity on local file system: %s GB, usage on remote file system: %s GB, margin: %s GB" % (local_capacity, remote_usage, margin_capacity))

    return local_capacity > remote_usage + margin_capacity / 2

def sync_fs_all(config, confgroup):
    """ Synchronize all file systems on this node with the partners of this node.

    This function creates the file systems to be failed over to this node with the
    same configuration on the paired nodes, except for namespace connectivity settings.
    Also this function deletes file systems that were deleted on the paired nodes
    or excluded from failover target.

    All created file systems are mounted. In other words, any file systems that
    are not mounted on the partner will not be created.

    No data restoration is executed.

    Args:
        config (Config.Config): Configuration of this script.
        confgroup (str): configuration group name of the pair configuration.
    Returns:
        None
    """

    # Load pair configuration
    print_status("Loading pair configuration ...")
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)
    stconf = FailoverConfig.StoragePolicyConfig(config.fo_conf_dir, confgroup)
    config.pairconf = pairconf

    # Setup connection to localhost
    # SSH credential is not required here because this is connection to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection("localhost", cred, config.retry_count, config.retry_interval, config.timeout)

    # Load HCP information
    print_status("Loading HCP configuration ...")
    try:
        hcpop = HDIOperation.HCPSetting(conn)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get HCP configuration on the node. host=%s, error=%s" % (conn.host, e.reason), e.code)

    # Find myself in the pair configuration
    print_status("Getting my IP address ...")
    self_ip = setup_common.get_my_ip_address(config, confgroup)
    if self_ip is None:
        print_status_error("Could not find my IP address in the pair configuration!")
        print_status("Please check the pair configuration if this node exists in the pair configuration.")
        raise Exceptions.ConfigurationError("This node is not defined in the pair configuration.")

    # Find my partners in the pair configuration
    my_pairs = pairconf.get_list(secondary=self_ip)
    log_info("Partners of %s: %s" % (self_ip, Utility.deep_str(my_pairs)))
    print_status("My partners: %s" % [pair.primary for pid, pair in my_pairs])

    # Prepare file system operation to this node
    fsop = HDIOperation.FileSystem(conn)

    # Get a list of file systems on this node
    print_status("Getting file systems on this node ...")
    try:
        fslist_on_this_node = fsop.get_all()
    except HDIOperation.OperationError, e:
        print_status_error("Failed to get a list of file systems on %s. error=%s" % (conn.host, str(e)))
        raise Exceptions.OperationError("Failed to get file system information. error=%s" % e.reason, e.code)

    # Download configuration of each partner
    fsdict_to_be_failed_over = {}
    target_fslist_all = []
    print_status("=== Download configuration of primary nodes ===")
    for pid, pair in my_pairs:
        # Download configuration file from HCP
        print_status("Downloading configuration of %s from HCP ..." % pair.primary)
        archive_path = backup_sync_common.download_configs(pair.primary, config.hdi_conf_dir, hcp_info, config)
        archive = HDIConfigArchive.HDIConfigArchive()
        archive.load(archive_path)

        # Get a list of file systems to be failed over
        if pair.is_all:
            # All file systems on the partner are to be failed over
            target_fslist = archive.fslist
        else:
            # Only selected file systems are to be failed over
            target_fslist = []
            for fsname in pair.target_fslist:
                target_fslist.extend([fsinfo for fsinfo in archive.fslist if fsinfo.fileSystemName == fsname])
        print_status("File systems to be failed over from %s: %s" % (pair.primary, [fsinfo.fileSystemName for fsinfo in target_fslist]))

        # Note: target_fslist includes a file system that is not supported
        # Not-supported file system will be skipped by sync_filesystem()
        target_fslist_all += [(pid, pair.primary, fsinfo) for fsinfo in target_fslist]
        fsdict_to_be_failed_over[pair.primary] = [fsinfo.fileSystemName for fsinfo in target_fslist]

    # Synchronize file systems in dictionary order of file system name
    target_fslist_all.sort(key=lambda e: e[2].fileSystemName)
    print_status("=== Synchronize file systems ===")
    for pid, primary, fsinfo in target_fslist_all:
        # Get storage policy for the file system
        policy = stconf.get_policy(pid, fsinfo.fileSystemName)
        if policy is None:
            log_error("Could not get any storage policy applicable to %s. pair_id=%s, primary=%s" % (fsinfo.fileSystemName, pid, primary))
            raise Exceptions.ConfigurationError("No applicable storage policy is defined for file system %s." % fsinfo.fileSystemName)

        log_info("Storage policy for %s: %s" % (fsinfo.fileSystemName, policy.to_display_string()))
        # Synchronize the file system
        print_status("--- Synchronize configuration of %s ---" % fsinfo.fileSystemName)
        try:
            sync_filesystem(fsinfo, primary, policy, config, fslist_on_this_node)
        except Exception, e:
            print_status_error("Synchronization of %s failed. error=%s" % (fsinfo.fileSystemName, str(e)))
            print_status("Continue synchronization of other file systems ...")
            global got_error
            got_error = True

def clean_up_unnecessary_filesystems(config, confgroup_list):
    ### Clean up file systems that don't exist on any partners
    print_status("=== Clean up unnecessary file systems on this node ===")

    # Get a list of file systems that have been synced on this node
    fsdict_synced = backup_sync_common.get_synced_fslist(config.hdi_conf_dir)
    for ipaddr in fsdict_synced.keys():
        print_status("File systems synced from %s: %s" % (ipaddr, fsdict_synced[ipaddr]))

    # Get a list of file systems that are supposed to be failed over to this node
    fsdict_to_be_failed_over = {}
    for confgroup in confgroup_list:
        # Load pair configuration
        print_status("Loading pair configuration ...")
        pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)

        # Find myself in the pair configuration
        print_status("Getting my IP address ...")
        self_ip = setup_common.get_my_ip_address(config, confgroup)
        if self_ip is None:
            print_status_error("Could not find my IP address in the pair configuration!")
            print_status("Please check the pair configuration if this node exists in the pair configuration.")
            raise Exceptions.ConfigurationError("This node is not defined in the pair configuration.")

        my_pairs = pairconf.get_list(secondary=self_ip)

        for pid, pair in my_pairs:
            # Get archive of the node. Assume the archive has been already downloaded.
            print_status("Loading configuration of %s ..." % pair.primary)
            archive_path = backup_sync_common.get_existing_archive_path(pair.primary, config.hdi_conf_dir)
            if archive_path is None:
                print_status_error("Configuration archive of %s is not found for some reason while it is expected to have been already downloaded." % pair.primary)
                raise Exceptions.ConfigurationError("Configuration archive of %s is not found." % pair.primary)
            archive = HDIConfigArchive.HDIConfigArchive()
            archive.load(archive_path)

            # Get a list of file systems to be failed over
            if pair.is_all:
                # All file systems on the partner are to be failed over
                target_fslist = archive.fslist
            else:
                # Only selected file systems are to be failed over
                target_fslist = []
                for fsname in pair.target_fslist:
                    target_fslist.extend([fsinfo for fsinfo in archive.fslist if fsinfo.fileSystemName == fsname])

            # Remove unsupported file systems from the list. These should not be considered
            # when deciding file systems to be cleaned up.
            target_fslist = [fsinfo for fsinfo in target_fslist if fsinfo.is_supported()]
            fsdict_to_be_failed_over[pair.primary] = [fsinfo.fileSystemName for fsinfo in target_fslist]

    for ipaddr in fsdict_to_be_failed_over.keys():
        print_status("File systems to be failed over from %s (supported): %s" % (ipaddr, fsdict_to_be_failed_over[ipaddr]))

    # Setup connection to localhost
    # SSH credential is not required here because this is connection to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection("localhost", cred, config.retry_count, config.retry_interval, config.timeout)

    # Prepare file system operation to this node
    fsop = HDIOperation.FileSystem(conn)

    # Detect and delete unnecessary file systems.
    # If a file system is synced on this node but is not included in failover target,
    # the file system is not necessary on this node and needs to be deleted.
    # This happens when:
    #     the file system was deleted on the primary node.
    #     the file system was removed from failover target.
    #     the primary node was removed from the pair with this node.
    for ipaddr in fsdict_synced.keys():
        if ipaddr in fsdict_to_be_failed_over.keys():
            # Compare the list of file systems synced from the primary node onto this node (A)
            # AND the list of file systems to be failed over from the primary node (B).
            # (A) - (B) is the file systems that can be deleted.
            fsset_synced = set(fsdict_synced[ipaddr])
            fsset_to_be_failed_over = set(fsdict_to_be_failed_over[ipaddr])
            fslist_to_be_deleted = list(fsset_synced - fsset_to_be_failed_over)
        else:
            # The primary node was synced but not included in failover target list.
            # All file system came from the node can be deleted.
            fslist_to_be_deleted = fsdict_synced[ipaddr]

        if len(fslist_to_be_deleted) == 0:
            print_status("No file systems from %s can be deleted." % ipaddr)
            continue

        print_status("These file systems from %s are not necessary on this node: %s" % (ipaddr, fslist_to_be_deleted))
        for fsname in fslist_to_be_deleted:
            try:
                print_status("Checking mount status of %s ..." % fsname)
                is_mounted = fsop.is_mounted(fsname)
                if is_mounted:
                    print_status("Unmounting %s ..." % fsname)
                    fsop.unmount(fsname)
                else:
                    print_status("%s is already unmounted." % fsname)
                print_status("Deleting %s ..." % fsname)
                fsop.delete(fsname)
                backup_sync_common.delete_fsinfo(fsname, ipaddr, config.hdi_conf_dir)
            except HDIOperation.OperationError, e:
                print_status_error("Failed to unmount/delete a file system %s on %s. error=%s" % (fsname, conn.host, str(e)))
                raise Exceptions.OperationError("Failed to delete a file system. error=%s" % e.reason, e.code)

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    short_opts = ""
    long_opts = ["check-resources", "send-alert"]
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    do_check_resources = False
    for o, a in opts:
        if o == "--check-resources":
            do_check_resources = True
        if o == "--send-alert":
            global send_alert
            send_alert = True

    ### Acquire execution lock
    print_status("Checking parallel execution ...")

    # Lock on self node
    execute_lock_path = lock_common.get_execute_lock_file()
    try:
        locker = open(execute_lock_path, 'w')
        fcntl.flock(locker, fcntl.LOCK_EX | fcntl.LOCK_NB)
        # This lock will be released when this process finishes
    except EnvironmentError, e:
        if e.errno == errno.EAGAIN:
            # The lock is already acquired by another process. Just give up here.
            print_status("File system synchronization or failover execution is already running. Cancel my execution.")
            raise Exceptions.NoExecution()
        else:
            log_error("An error occurred while getting a lock for execution. error=%s" % str(e))
            raise
    log_info("Execution lock has been acquired.")

    # Wait if execution lock is acquired on the other node in cluster
    lock_common.check_execution_lock_on_the_other_node(config)

    # Check resource group status
    print_status("Checking resource group status ...")
    if not getinfo_common.is_running_healthily(config):
        print_status_warning("The node is not running healthily. Stop execution.")
        raise Exceptions.ConfigurationError("Node status is not healthy.")

    # Check resources before synchronizing file systems.
    # Do not get backup on the primary nodes because it should be already done
    # by regular execution of backup_configs.
    if do_check_resources:
        print_status("Execute checking resources.")
        myself = setup_common.get_my_ip_address(config)
        result = check_resources.check_resources(myself, config, nobackup=True)
        if result['has_conflict'] or result['has_shortage']:
            print_status_warning("There is resource name conflict or storage shortage.")
            if send_alert:
                Alert.send_alert(Alert.MsgCode.RESOURCE_ERROR, detail="See logs or execute check_resources to see details")
        if result['is_MQE_unavailable']:
            print_status_warning("MQE is not available or indexing is not enabled.")
            if send_alert:
                Alert.send_alert(Alert.MsgCode.MQE_ERROR)
        # Continue to synchronize file systems anyway.
        # There might be a situation that conflict or shortage is expected by the admin.

    # Load the existing pair configuration
    log_info("Getting list of pair configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Synchronize all file systems on each Configuration Group
    for confgroup in sorted(confgroup_list):
        sync_fs_all(config, confgroup)

    # Clean up unnecessary file systems that were created on this node
    clean_up_unnecessary_filesystems(config, sorted(confgroup_list))

    print_status("Completed.")

    # Raise an error if we got an error on file system synchronization
    if got_error:
        raise Exceptions.OperationError("An error occurred while synchronizing a file system.", -1)

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_OK)
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail=str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail="Interrupted by a signal.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail="Unable to load script configuration file.")
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail=str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail=str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail=str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail=str(e))
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        if send_alert:
            Alert.send_alert(Alert.MsgCode.SYNC_ERROR, detail=str(e))
        sys.exit(99)

    if send_alert:
        Alert.send_alert(Alert.MsgCode.SYNC_OK)
    sys.exit(0)
